// エンコーディング：UTF-8
module Kun;

import KunCode;
import Numbered;
import std.file;
import std.stdio;
import std.stream;
import std.string;
import TimeName;

class Kun {
    const bool COMMENTS = true;
    const bool SERIAL_NUMBER = true;
    const bool WITHOUT_COMMENTS = false;
    const byte BYTE_DOT = 0x2E; // '.'
    const byte BYTE_0 = 0x30; // '0'
    const byte BYTE_1 = 0x31; // '1'
    const byte BYTE_K = 0x4B; // 'K'
    const byte BYTE_N = 0x4E; // 'N'
    const byte BYTE_U = 0x55; // 'U'
    const int ALL = 37;
    const int BITS = 8;
    const int CELL_HEIGHT = 4;
    const int CELL_WIDTH = 4;
    const int CSV_SPECIFIED = 0;
    const int FROM_COMMA_SIZE = 5;
    const int ITEM_COUNT = 30;
    const int ITEM_WIDTH = 8;
    const int KOPIC_FROM = 400; // 0 < KOPIC_FROM
    const int KOPIC_KUN_SIZE = 1688; // ( XBM_WIDTH + 2 ) * XBM_HEIGHT + SIGNATURE
    const int KOPIC_REPEAT = 420; // KOPIC_FROM < KOPIC_REPEAT
    const int KUN_EXPAND = 1;
    const int KUN_FILE_ERROR = -1;
    const int KUN_SHRINK = 2;
    const int LOG_COMMA_SIZE = 4;
    const int NOT_TREATED_AS_TRUE = 3;
    const int RECOVERY = 4;
    const int REPEAT_COMMA_SIZE = 7;
    const int SETTING_COMMA_SIZE = 8;
    const int SHARP_TEXT = 5;
    const int SIGNATURE = 8;
    const int TREATED_AS_FALSE = 6;
    const int VERSION_COMMENTS = 8;
    const int VERSION_GUIDE = 7;
    const int XBM_FILE_ERROR = 9;
    const int XBM_HEIGHT = 40;
    const int XBM_HEIGHT_1 = XBM_HEIGHT + 1;
    const int XBM_READ = 10;
    const int XBM_TO_CSV = 11;
    const int XBM_WIDTH = 40;
    const int XBM_HEIGHT_XBM_WIDTH = XBM_HEIGHT * XBM_WIDTH;
    const int XBM_HEIGHT_XBM_WIDTH_BITS = XBM_HEIGHT * XBM_WIDTH/BITS;
    const char[] CARET = "^";
    const char[] COMMA = ",";
    const char[] CSV_DECREMENT = "-cd";
    const char[] CSV_FILE = "-c";
    const char[] EASY_TO_SEE = "-t";
    const char[] EMPTY = "";
    const char[] ERROR_TEXT = "error.txt";
    const char[] FALSE_TEXT = "false.txt";
    const char[] FROM_COMMA = "From,";
    const char[] FROM_NOT_38 = "-38-";
    const char[] KOPIC_CSV = "kopic.csv";
    const char[] KOPIC_KUN = "kopic.kun";
    const char[] KOPIC_TEXT = "kopic.txt";
    const char[] KOPIC_XBM = "kopic.xbm";
    const char[] KUN_FILE = "-k";
    const char[] LOG_COMMA = "Log,";
    const char[] MANAGE_TEXT = "manage.txt";
    const char[] N_0TO9 = "0123456789";
    const char[] NEW_XBM = "-n";
    const char[] NO_REMOVING = "-r-";
    const char[] QUOTE = "\"";
    const char[] REMOVE_ERROR_TEXT = "remove_e.txt";
    const char[] REPEAT_COMMA = "Repeat,";
    const char[] SETTING_COMMA = "Setting,";
    const char[] SHARP = "#";
    const char[] SIZE = "-s";
    const char[] SPACE = " ";
    const char[] TEXT_FILE = ".txt";
    const char[] TRU = "tru";
    const char[] TRUE_TEXT = "true.txt";
    const char[] version_comments = "-VC";
    const char[] version_guide = "-V";
    const char[] VERSION_TEXT = "version.txt";
    const char[] XBM_RECOVERY = "-y";
    const char[] ZERO_X = "0x";
    const size_t SIGNATURE_SIZE = SIGNATURE;
    const ubyte DEL = 0x7F;

    bool check;
    bool resize = false;
    bool[XBM_WIDTH][XBM_HEIGHT] mirror_images;
    bool[XBM_WIDTH][XBM_HEIGHT] xbm_bool_array;
    byte black, white;
    byte[KOPIC_KUN_SIZE] run_length;
    char[] error, file_name;
    char[][] h = ["00", "01", "02", "03", "04", "05", "06", "07", "08", "09", "0a", "0b", "0c", "0d", "0e", "0f", "10", "11", "12", "13", "14", "15", "16", "17", "18", "19", "1a", "1b", "1c", "1d", "1e", "1f", "20", "21", "22", "23", "24", "25", "26", "27", "28", "29", "2a", "2b", "2c", "2d", "2e", "2f", "30", "31", "32", "33", "34", "35", "36", "37", "38", "39", "3a", "3b", "3c", "3d", "3e", "3f", "40", "41", "42", "43", "44", "45", "46", "47", "48", "49", "4a", "4b", "4c", "4d", "4e", "4f", "50", "51", "52", "53", "54", "55", "56", "57", "58", "59", "5a", "5b", "5c", "5d", "5e", "5f", "60", "61", "62", "63", "64", "65", "66", "67", "68", "69", "6a", "6b", "6c", "6d", "6e", "6f", "70", "71", "72", "73", "74", "75", "76", "77", "78", "79", "7a", "7b", "7c", "7d", "7e", "7f", "80", "81", "82", "83", "84", "85", "86", "87", "88", "89", "8a", "8b", "8c", "8d", "8e", "8f", "90", "91", "92", "93", "94", "95", "96", "97", "98", "99", "9a", "9b", "9c", "9d", "9e", "9f", "a0", "a1", "a2", "a3", "a4", "a5", "a6", "a7", "a8", "a9", "aa", "ab", "ac", "ad", "ae", "af", "b0", "b1", "b2", "b3", "b4", "b5", "b6", "b7", "b8", "b9", "ba", "bb", "bc", "bd", "be", "bf", "c0", "c1", "c2", "c3", "c4", "c5", "c6", "c7", "c8", "c9", "ca", "cb", "cc", "cd", "ce", "cf", "d0", "d1", "d2", "d3", "d4", "d5", "d6", "d7", "d8", "d9", "da", "db", "dc", "dd", "de", "df", "e0", "e1", "e2", "e3", "e4", "e5", "e6", "e7", "e8", "e9", "ea", "eb", "ec", "ed", "ee", "ef", "f0", "f1", "f2", "f3", "f4", "f5", "f6", "f7", "f8", "f9", "fa", "fb", "fc", "fd", "fe", "ff"];
    int cell_height = CELL_HEIGHT;
    int cell_width = CELL_WIDTH;
    int code, data_height, data_width, decrement_by_n, encoding_size;
    int[ char[] ] hexadecimal_to_integer;
    int[XBM_HEIGHT_XBM_WIDTH_BITS] data_bits;
    KunCode kc;
    uint[] shift = [
        0b00000001,
        0b00000010,
        0b00000100,
        0b00001000,
        0b00010000,
        0b00100000,
        0b01000000,
        0b10000000
    ];

    private bool createErrorText(char[] messages) {
        try {
            Stream file = new BufferedFile(ERROR_TEXT, FileMode.OutNew);
            file.writefln(messages);
            file.close();
            return true;
        } catch(OpenException e34) {
            error = "OpenException e34";
            return false;
        }

    }

    private bool createFalseText() {
        try {
            Stream file = new BufferedFile(FALSE_TEXT, FileMode.OutNew);
            file.writefln(" ");
            file.writefln("0");
            file.writefln("　"); // UTF-8
            file.close();
            return true;
        } catch(OpenException e33) {
            error = "OpenException e33";
            return false;
        }

    }

    private bool createKopicXbm() {
        try {
            Stream file = new BufferedFile(KOPIC_XBM, FileMode.OutNew);

            file.writef("#define data_width ");
            file.writefln(XBM_WIDTH);
            file.writef( "#define data_height ");
            file.writefln(XBM_HEIGHT);
            file.writefln("static char data_bits[] = {");
            file.writefln("0x00,0x00,0x40,0x00,0x00,");
            file.writefln("0x00,0x00,0x80,0x00,0x00,");
            file.writefln("0x00,0x00,0x00,0x01,0x00,");
            file.writefln("0x00,0x00,0x00,0x02,0x00,");
            file.writefln("0x00,0x00,0x00,0x02,0x00,");
            file.writefln("0x00,0x00,0x00,0x04,0x00,");
            file.writefln("0x00,0x00,0x00,0x04,0x00,");
            file.writefln("0x00,0x00,0x00,0x04,0x00,");
            file.writefln("0x00,0x00,0x00,0x04,0x00,");
            file.writefln("0x00,0x00,0x00,0x04,0x00,");
            file.writefln("0x00,0x00,0x00,0x02,0x00,");
            file.writefln("0x00,0x00,0x00,0x02,0x00,");
            file.writefln("0x00,0x00,0x00,0x02,0x00,");
            file.writefln("0x00,0x00,0x00,0x01,0x00,");
            file.writefln("0x00,0x00,0x00,0x01,0x00,");
            file.writefln("0x00,0x00,0x80,0x00,0x00,");
            file.writefln("0x00,0x00,0x80,0x00,0x00,");
            file.writefln("0x00,0x00,0x40,0x00,0x00,");
            file.writefln("0x00,0x00,0x20,0x00,0x00,");
            file.writefln("0x00,0x00,0x20,0x00,0x00,");
            file.writefln("0x00,0x00,0x10,0x00,0x00,");
            file.writefln("0x00,0x00,0x08,0x00,0x00,");
            file.writefln("0x00,0x00,0x08,0x00,0x00,");
            file.writefln("0x00,0x00,0x04,0x00,0x00,");
            file.writefln("0x00,0x00,0x02,0x00,0x00,");
            file.writefln("0x00,0x00,0x02,0x00,0x00,");
            file.writefln("0x00,0x00,0x01,0x00,0x00,");
            file.writefln("0x00,0x00,0x01,0x00,0x00,");
            file.writefln("0x00,0x80,0x00,0x00,0x00,");
            file.writefln("0x00,0x80,0x00,0x00,0x00,");
            file.writefln("0x00,0x80,0x00,0x00,0x00,");
            file.writefln("0x00,0x80,0x00,0x00,0x00,");
            file.writefln("0x00,0x80,0x00,0x00,0x00,");
            file.writefln("0x00,0x40,0x00,0x00,0x00,");
            file.writefln("0x00,0x80,0x00,0x00,0x00,");
            file.writefln("0x00,0x00,0x01,0x00,0x00,");
            file.writefln("0x00,0x00,0x01,0x00,0x00,");
            file.writefln("0x00,0x00,0x02,0x00,0x00,");
            file.writefln("0x00,0x00,0x0c,0x00,0x00,");
            file.writefln("0x00,0x00,0x10,0x00,0x00");
            file.writefln("};");

            file.close();
            return true;
        } catch(OpenException e26) {
            error = "OpenException e26";
            return false;
        }

    }

    private bool createManageText() {
        try {
            Stream file = new BufferedFile(MANAGE_TEXT, FileMode.OutNew);
            file.writefln("0");
            file.writefln("4");
            file.writefln("4");
            file.close();
            return true;
        } catch(OpenException e25) {
            error = "OpenException e25";
            return false;
        }

    }

    private bool createTrueText() {
        try {
            Stream file = new BufferedFile(TRUE_TEXT, FileMode.OutNew);
            file.writefln("*");
            file.writefln("■"); // UTF-8
            file.close();
            return true;
        } catch(OpenException e6) {
            error = "OpenException e6";
            return false;
        }

    }

    private bool itemsTreatedAsFalse(char[] file_name) {
// Software replaces true characters with quotes beforehand.
        char[] s;
        char[][] items = ["",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  ""];

        try {
// Acquiring items from FALSE_TEXT so that items can be treated as false.
            Stream buffered = new BufferedFile(FALSE_TEXT); 
            for(int k = 0; k < ITEM_COUNT; k++) {
                s = buffered.readLine();
                if(s == null) {
                    break;
                }

                items[k] = s;
            }

            buffered.close();
        } catch(OpenException e35) {
            error = "OpenException e35";
            createErrorText(FALSE_TEXT);
            return false;
        } catch(ReadException e15) {
            error = "OpenException e15";
            createErrorText(FALSE_TEXT);
            return false;
        }

        try {
            Stream text = new BufferedFile(file_name);
            int h, item_size, q, s_length, z;
            int w = 0;
            char[] c, head, item, tail;
            char[] data = EMPTY;

            for(h = 0; ; ++h) {
                s = text.readLine();
                if(s == null) {
                    break;
                } if( s.length == 0 ) {
                    break;
                }

// Software replaces false characters with carets.
                for(int k = 0; k < ITEM_COUNT; ++k) {
                    item = items[k];
                    item_size = item.length;
                    if(item_size == 0) {
                        break;
                    }

                    data = EMPTY;
                    tail = s;
                    for(;;) {
                        q = find(tail, item);
                        if(q >= 0) {
                            head = tail[0..q];
                            q += item_size;
                            tail = tail[q..$];
                            data ~= head;
                            data ~= CARET;
                        } else {
                            data ~= tail;
                            break;
                        }

                    }

                    s = data;
                }

// Software treats carets as false.
                s_length = s.length; // new length
                w = 0;
                for(int k = 0; k < s_length; k++) {
                    z = k + 1;
                    c = s[k..z]; // c is assumed to be ASCII because software replaces true characters with quotes beforehand.
                    if(CARET == c) {
                        xbm_bool_array[h][w++] = false;
                    } else {
                        xbm_bool_array[h][w++] = true;
                    }

                }

            }

            text.close();
            data_height = h;
            data_width = w;
            decrement_by_n = 0;
            rewrite_xbm(KOPIC_XBM, data_width, data_height);
            return true;
        } catch(OpenException e36) {
            error = "OpenException e36";
            createErrorText(file_name);
            return false;
        } catch(ReadException e38) {
            error = "ReadException e38";
            createErrorText(file_name);
            return false;
        }

    }

    private bool itemsTreatedAsTrue(char[] file_name) {
        char[] os = EMPTY;
        char[] s;
        char[][] items = ["",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  "",  ""];

        try {
// Acquiring items from TRUE_TEXT so that items can be treated as true.
            Stream buffered = new BufferedFile(TRUE_TEXT); 
            for(int k = 0; k < ITEM_COUNT; k++) {
                s = buffered.readLine();
                if(s == null) {
                    break;
                }

                items[k] = s;
            }

            buffered.close();
        } catch(OpenException e5) {
            error = "OpenException e5";
            createErrorText(TRUE_TEXT);
            return false;
        } catch(ReadException e4) {
            error = "OpenException e4";
            createErrorText(TRUE_TEXT);
            return false;
        }

        try {
            Stream text = new BufferedFile(file_name); 
            int h, item_size, q, s_length, z;
            int w = 0;
            char[] c, head, item, tail;
            char[] data = EMPTY;

            for(h = 0; ; ++h) {
                s = text.readLine();
                if(s == null) {
                    break;
                }

// Software replaces true characters with quotes.
                for(int k = 0; k < ITEM_COUNT; ++k) {
                    item = items[k];
                    item_size = item.length;
                    if(item_size == 0) {
                        break;
                    }

                    data = EMPTY;
                    tail = s;
                    for(;;) {
                        q = find(tail, item);
                        if(q >= 0) {
                            head = tail[0..q];
                            q += item_size;
                            tail = tail[q..$];
                            data ~= head;
                            data ~= QUOTE;
                        } else {
                            data ~= tail;
                            break;
                        }

                    }

                    s = data;
                }

                os ~= s;
                os ~= newline;
            }

            text.close();
        } catch(OpenException e3) {
            error = "OpenException e3";
            createErrorText(file_name);
            return false;
        } catch(ReadException e2) {
            error = "ReadException e2";
            createErrorText(file_name);
            return false;
        }

        TimeName tt = new TimeName(TRU, TEXT_FILE);
        kc.trueKopicText = tt.temporaryName(SERIAL_NUMBER);
        try {
            Stream file = new BufferedFile(kc.trueKopicText, FileMode.OutNew);
            file.writef(os);
            file.close();
            return true;
        } catch(OpenException e20) {
            error = "OpenException e20";
            return false;
        }

    }

    private bool kunExpand(char[] file_name) {
        bool[XBM_HEIGHT_XBM_WIDTH] expanded;
        const int BLACK = 1;
        const int WHITE = 0;
        int black_or_white, count, DEL_count, n, q, sum, u;
        int[XBM_HEIGHT_1] data_width;
        ulong file_size;

        try {
            file_size = getSize(file_name);
        } catch(FileException e22) {
            error = "FileException e22";
            return false;
        }

        ubyte[] buffer = new ubyte[file_size];

        try {
            Stream buffered = new BufferedFile(file_name); 
            size_t length = buffered.read(buffer);
            if(length < SIGNATURE_SIZE) {
                buffered.close();
                return false;
            }

            buffered.close();
            n = 0;
            DEL_count = 0;
            q = 0;
            sum = 0;
            for(int k = SIGNATURE; k < file_size && DEL_count < XBM_HEIGHT_1; ++k, ++q) {
                if(buffer[k] == DEL) {
                    data_width[DEL_count] = sum;
                    DEL_count += 1;
                    q = 0;
                    sum = 0;
                    continue;
                }

                count = buffer[k];
                sum += count;
                black_or_white = q%2;
                switch(black_or_white) {
                case BLACK:
                    for(u = 0; u < count; u++) {
                        expanded[n++] = true;
                    }

                    break;
                case WHITE:
                    for(u = 0; u < count; u++) {
                        expanded[n++] = false;
                    }

                    break;
                default:
                    break;
                }

            }

// Checking if data width is constant.
            data_height = DEL_count - 1;
            for(int m = 2; m < data_height; ++m) {
                if(data_width[m] == data_width[1]) {
                    continue;
                } else {
                    char[] os = EMPTY;
                    os ~= "data_width[1] = ";
                    os ~= std.string.toString(data_width[1]);
                    os ~= newline;
                    os ~= "data_width[";
                    os ~= std.string.toString(m);
                    os ~= "] = ";
                    os ~= std.string.toString(data_width[m]);
                    createErrorText(os);
                    return false;
                }

            }

        } catch(OpenException e13) {
            error = "OpenException e13";
            createErrorText(file_name);
            return false;
        } catch(ReadException e37) {
            error = "ReadException e37";
            createErrorText(file_name);
            return false;
        }

        try {
            Stream file = new BufferedFile(KOPIC_TEXT, FileMode.OutNew);
            n = 0;
            for(int m = 0; m < data_height; ++m) {
                for(q = 0; q < data_width[1]; ++q) {
                    xbm_bool_array[m][q] = expanded[n++];

                    if(xbm_bool_array[m][q]) {
                        file.writef("#");
                    } else {
                        file.writef(" ");
                    }

                }

                file.writefln("");
            }

            file.close();
            return true;
        } catch(OpenException e39) {
            error = "OpenException e39";
            createErrorText(KOPIC_TEXT);
            return false;
        }

    }

    private bool rewrite_xbm(char[] fileName, int data_width, int data_height) {
        int DATA_HEIGHT_1 = data_height - 1;
        int ITEM_WIDTH_1 = ITEM_WIDTH - 1;
        int q;
        uint b;

        mirrorImages(data_width, data_height);

        try {
            Stream file = new BufferedFile(fileName, FileMode.OutNew);
            file.writef("#define data_width ");
            file.writefln(data_width);
            file.writef("#define data_height ");
            file.writefln(data_height);
            file.writefln("static char data_bits[] = {");

            for(int m = 0; m < data_height; ++m) {
                q = 0;
                for(int k = 0; k < data_width; k += ITEM_WIDTH) {
                    b = 0;

                    for(int p = ITEM_WIDTH_1; p >= 0; --p, ++q) {
                        if(mirror_images[m][q]) {
                            b |= shift[p];
                        }

                    }

                    if(m == DATA_HEIGHT_1 && q == data_width) {
                        file.writef(ZERO_X);
                        file.writef(h[b]);
                    } else {
                        file.writef(ZERO_X);
                        file.writef(h[b]);
                        file.writef(",");
                    }

                }

                file.writefln("");
            }

            file.writefln("};");
            file.close();
            return true;
        } catch(OpenException e29) {
            error = "OpenException e29";
            createErrorText(file_name);
            return false;
        }

    }

    private bool sharp_text(char[] fileName, int data_width, int data_height) {
        try {
            Stream file = new BufferedFile(fileName, FileMode.OutNew);
            for(int m = 0; m < data_height; ++m) {
                for(int q = 0; q < data_width; ++q) {
                    if(xbm_bool_array[m][q]) {
                        file.writef(SHARP);
                    } else {
                        file.writef(SPACE);
                    }

                }

                file.writefln("");
            }

            file.close();
            return true;
        } catch(OpenException e30) {
            error = "OpenException e30";
            textFailure(fileName);
            return false;
        }

    }

    private bool textFailure(char[] fileName) {
        try {
            Stream file = new BufferedFile(ERROR_TEXT, FileMode.OutNew);
            file.writef(fileName);
            file.close();
            return true;
        } catch(OpenException e31) {
            error = "OpenException e31";
            return false;
        }

    }

    private int checkExtension(char[] file_name) {
        char[] after_dot;
        const int CSV = 0;
        const int KUN = 1;
        const int KOPIC_EXTENSION_TEXT = 2;
        const int XBM = 3;
        int extension = KOPIC_EXTENSION_TEXT;

        int q = find(file_name, ".");
        if(q >= 0) {
            after_dot = file_name[q..$];
        } else {
            return extension;
        }

        if(after_dot == ".csv") {
            extension = CSV;
        } else if(after_dot == ".xbm") {
            extension = XBM;
        } else if(after_dot == ".kun") {
            extension = KUN;
        }

        return extension;
    }

    private int createKopicCsv(char[] essential, int data_width, int data_height, bool from_not_38) {
        for(int i = 0; i < KOPIC_KUN_SIZE; ++i) {
            run_length[] = 0;
        }

        int m = 0;
        encoding_size = 0;

        run_length[encoding_size] = DEL;
        encoding_size += 1;
        counting(data_width, m);

        for(m = 1; m < data_height; ++m) {
            if(check) {
                run_length[encoding_size] = black;
            } else {
                run_length[encoding_size] = white;
            }

            encoding_size += 1;

            run_length[encoding_size] = DEL;
            encoding_size += 1;
            counting(data_width, m);
        }

        if(check) {
            run_length[encoding_size] = black;
        } else {
            run_length[encoding_size] = white;
        }

        encoding_size += 1;

        run_length[encoding_size] = DEL;
        encoding_size += 1;

        try {
            Stream file = new BufferedFile(KOPIC_CSV, FileMode.OutNew);
            char[][] d = ["", "01,", "02,", "03,", "04,", "05,", "06,", "07,", "08,", "09,", "10,", "11,", "12,", "13,", "14,", "15,", "16,", "17,", "18,", "19,", "20,", "21,", "22,", "23,", "24,", "25,", "26,", "27,", "28,", "29,", "30,", "31,", "32,", "33,", "34,", "35,", "36,", "37,", "38,"];

            if(from_not_38) { // Use From, not 38.
                d[38] = "From,";
            }

            int D_ARRAY_SIZE = ALL + 1;
            int count, value;
            int DEL_count = 0;
            int sum = 0;
            for(int k = 0; k < encoding_size && DEL_count < D_ARRAY_SIZE; ++k) {
                if(run_length[k] == DEL) {
                    DEL_count += 1;
                    if(sum > 0) {
                        file.writefln("");
                    }

                    file.writef(d[DEL_count]);
                    sum = 0;

// After run_length[k] equals DEL, software omits the first zero.
                    ++k;
                    if( k < encoding_size && DEL_count < D_ARRAY_SIZE ) {
                        ++k;
                        if( k < encoding_size && DEL_count < D_ARRAY_SIZE ) {
// After omitting the first zero, software writes a value.
                            count = run_length[k];
                            sum += count;
                            value = count - decrement_by_n;
                            if(value < 0) {
                                value = 0;
                            }

                            file.writef(value);
                            continue;
                        }

                    }

                }

                count = run_length[k];
                sum += count;
            }

            file.writef(essential);
            file.close();
            return encoding_size;
        } catch(OpenException e32) {
            error = "OpenException e32";
            return KUN_FILE_ERROR;
        }

    }

    private int sequence_of_non_numbers(char[] line) {
        int q = 0;
        int size = line.length;
        int z;
        char[] c;
        for(int i = 0; i < size; ++i) {
            z = i + 1;
            c = line[i..z];
            q = find(N_0TO9, c);
            if(q >= 0) {
                q = i;
                break;
            } else {
                continue;
            }

        }

        return q;
    }

    private int sequenceOfNumbers(char[] line) {
        int q = 0;
        int size = line.length;
        int z;
        char[] c;
        for(int i = 0; i < size; ++i) {
            z = i + 1;
            c = line[i..z];
            q = find(N_0TO9, c);
            if(q >= 0) {
                continue;
            } else {
                q = i;
                break;
            }

        }

        return q;
    }

    private int shrink(int data_width, int data_height) {
        int m = 0;
        encoding_size = 0;

        run_length[encoding_size] = DEL;
        encoding_size += 1;
        counting(data_width, m);

        for(m = 1; m < data_height; ++m) {
            if(check) {
                run_length[encoding_size] = black;
            } else {
                run_length[encoding_size] = white;
            }

            encoding_size += 1;

            run_length[encoding_size] = DEL;
            encoding_size += 1;
            counting(data_width, m);
        }

        if(check) {
            run_length[encoding_size] = black;
        } else {
            run_length[encoding_size] = white;
        }

        encoding_size += 1;

        run_length[encoding_size] = DEL;
        encoding_size += 1;

        try {
            Stream binary = new BufferedFile(KOPIC_KUN, FileMode.OutNew);
            byte byte_height = cast(byte)(0xFF & data_height);
            byte byte_width = cast(byte)(0xFF & data_width);

// signature of 8 bytes
            binary.write(BYTE_DOT);
            binary.write(BYTE_K);
            binary.write(BYTE_U);
            binary.write(BYTE_N);
            binary.write(byte_height);
            binary.write(byte_width);
            binary.write(BYTE_1);
            binary.write(BYTE_0);

            for(int i = 0; i < encoding_size; ++i) {
                binary.write(run_length[i]);
            }

            binary.close();
            return encoding_size;
        } catch(OpenException e14) {
            error = "OpenException e14";
            createErrorText(error);
            return KUN_FILE_ERROR;
        }

    }

    private char[] inheritTerms() {
        bool useFrom = false;
        char[] term;
        char[] os = EMPTY;
        try {
            Stream buffered = new BufferedFile(KOPIC_CSV); 
            char[] s;
            int q;
            for(;;) {
                s = buffered.readLine();
                if(s == null) {
                    break;
                }

                q = find(s, FROM_COMMA);
                if(q >= 0) {
                    useFrom = true;
                    q += FROM_COMMA_SIZE;
                    term = s[q..$];
                    os ~= term;
                    os ~= newline;
                    continue;
                }

                q = find(s, LOG_COMMA);
                if(q >= 0) {
                    q += LOG_COMMA_SIZE;
                    term = s[q..$];
                    os ~= LOG_COMMA;
                    os ~= term;
                    os ~= newline;
                    continue;
                }

                q = find(s, REPEAT_COMMA);
                if(q >= 0) {
                    q += REPEAT_COMMA_SIZE;
                    term = s[q..$];
                    os ~= REPEAT_COMMA;
                    os ~= term;
                    os ~= newline;
                    continue;
                }

                q = find(s, SETTING_COMMA);
                if(q >= 0) {
                    q += SETTING_COMMA_SIZE;
                    term = s[q..$];
                    os ~= SETTING_COMMA;
                    os ~= term;
                    os ~= newline;
                    continue;
                }

            }

            buffered.close();

            char[] fourLines;
            if(useFrom) {
                fourLines = os;
            } else {
                fourLines = "0" ~ newline ~ os;
            }

            return fourLines;
        } catch(OpenException e16) {
            error = "OpenException e16";
            createErrorText(KOPIC_CSV);
            char[] fourLines = standardTerms();
            return fourLines;
        } catch(ReadException e21) {
            error = "OpenException e21";
            createErrorText(KOPIC_CSV);
            char[] fourLines = standardTerms();
            return fourLines;
        }

    }

    private char[] standardTerms() {
        char[] os = EMPTY;
        os ~= "0"; os ~= newline;
        os ~= "Log,KO.LOG"; os ~= newline;
        os ~= "Repeat,420"; os ~= newline;
        os ~= "Setting,123456"; os ~= newline;
        return os;
    }

    private void counting(int data_width, int m) {
        check = true;
        black = 0;
        white = 0;

        for(int q = 0; q < data_width; ++q) {
            if(check) {
                if(xbm_bool_array[m][q] == check) {
                    black += 1;
                } else {
                    run_length[encoding_size] = black;
                    encoding_size += 1;
                    check = xbm_bool_array[m][q];
                    white = 1;
                }

            } else {
                if(xbm_bool_array[m][q] == check) {
                    white += 1;
                } else {
                    run_length[encoding_size] = white;
                    encoding_size += 1;
                    check = xbm_bool_array[m][q];
                    black = 1;
                }

            }

        }

    }

    private void guide(bool comments) {
        const double VERSION = 1.02;

        try {
            Stream file = new BufferedFile(VERSION_TEXT, FileMode.OutNew);

            if(comments) {
                file.writef("viewXBM DFL Version ");
                file.writefln(VERSION);
                file.writefln("");
                file.writefln("viewXBM         # Read kopic.xbm ");
                file.writefln("viewXBM -c      # Produce kopic.csv from kopic.xbm ");
                file.writefln("viewXBM -c -38- # Use From, not 38. ");
                file.writefln("viewXBM -cd2    # Produce kopic.csv from kopic.xbm and decrement by 2.");
                file.writefln("viewXBM -k      # Produce kopic.kun from kopic.xbm ");
                file.writefln("viewXBM -r-     # Do not remove temporary files. ");
                file.writefln("viewXBM -s9-9   # Every cell is 9 pixels wide and 9 pixels high. ");
                file.writefln("viewXBM -t      # Produce kopic.txt from kopic.xbm ");
                file.writefln("viewXBM -y      # recovery ");
                file.writefln("viewXBM -V      # version ");
                file.writefln("viewXBM -VC     # version with comments ");
                file.writefln("viewXBM file_name.kun    # Produce kopic.txt from file_name.kun ");
                file.writefln("viewXBM file_name.txt    # Produce kopic.xbm from file_name.txt ");
                file.writefln("viewXBM file_name.xbm    # Read file_name.xbm ");
                file.writefln("viewXBM file_name.txt -n # Produce and read new kopic.xbm ");
            } else {
                file.writef("viewXBM DFL Version ");
                file.writefln(VERSION);
                file.writefln("");
                file.writefln("viewXBM         ");
                file.writefln("viewXBM -c      ");
                file.writefln("viewXBM -c -38- ");
                file.writefln("viewXBM -cd2    ");
                file.writefln("viewXBM -k      ");
                file.writefln("viewXBM -r-     ");
                file.writefln("viewXBM -s9-9   ");
                file.writefln("viewXBM -t      ");
                file.writefln("viewXBM -y      ");
                file.writefln("viewXBM -V      ");
                file.writefln("viewXBM -VC     ");
                file.writefln("viewXBM file_name.kun    ");
                file.writefln("viewXBM file_name.txt    ");
                file.writefln("viewXBM file_name.xbm    ");
                file.writefln("viewXBM file_name.txt -n ");
            }

            file.close();
        } catch(OpenException e40) {
            error = "OpenException e40";
        }

    }

    private void mirrorImages(int data_width, int data_height) {
        int q, r;

        for(int m = 0; m < data_height; ++m) {
            q = 0;
            for(int k = 0; k < data_width; k += ITEM_WIDTH) {
                r = k + ITEM_WIDTH - 1;
                mirror_images[m][r--] = xbm_bool_array[m][q++];
                mirror_images[m][r--] = xbm_bool_array[m][q++];
                mirror_images[m][r--] = xbm_bool_array[m][q++];
                mirror_images[m][r--] = xbm_bool_array[m][q++];
                mirror_images[m][r--] = xbm_bool_array[m][q++];
                mirror_images[m][r--] = xbm_bool_array[m][q++];
                mirror_images[m][r--] = xbm_bool_array[m][q++];
                mirror_images[m][r] = xbm_bool_array[m][q++];
            }

        }

    }

    private void removeTemporaryFiles() {
// Removing temporary files
        try {
            Stream buffered = new BufferedFile(REMOVE_ERROR_TEXT); 
            char[] s;
            for(;;) {
                s = buffered.readLine();
                if(s == null) {
                    break;
                } else if(s == TRUE_TEXT) {
                    continue;
                }

                std.file.remove(s);
            }

            buffered.close();
            std.file.remove(REMOVE_ERROR_TEXT);
        } catch(OpenException e17) {
            error = "OpenException e17";
            createErrorText(REMOVE_ERROR_TEXT);
        } catch(ReadException e12) {
            error = "OpenException e12";
            createErrorText(REMOVE_ERROR_TEXT);
        } catch(FileException e27) {
            error = "OpenException e27";
            createErrorText(REMOVE_ERROR_TEXT);
        }

    }

    private void xbmBooleanArray(int data_width, int data_height) {
        int q;
        int n = 0; // data_bits[n]
        uint b, s;

        for(int m = 0; m < data_height; ++m) {
            q = 0;
            for(int k = 0; k < data_width; k += ITEM_WIDTH, ++n) {
                b = data_bits[n];

                for(int p = 0; p < ITEM_WIDTH; ++p) {
                    s = b & shift[p];
                    if(s == shift[p]) {
                        xbm_bool_array[m][q++] = true;
                    } else {
                        xbm_bool_array[m][q++] = false;
                    }

                }

            }

        }

    }

    this(int d) {
        decrement_by_n = d;
        hexadecimal_to_integer["0x00"] = 0; hexadecimal_to_integer["0x01"] = 1; hexadecimal_to_integer["0x02"] = 2; hexadecimal_to_integer["0x03"] = 3; hexadecimal_to_integer["0x04"] = 4; hexadecimal_to_integer["0x05"] = 5; hexadecimal_to_integer["0x06"] = 6; hexadecimal_to_integer["0x07"] = 7; hexadecimal_to_integer["0x08"] = 8; hexadecimal_to_integer["0x09"] = 9;
        hexadecimal_to_integer["0x0a"] = 10; hexadecimal_to_integer["0x0b"] = 11; hexadecimal_to_integer["0x0c"] = 12; hexadecimal_to_integer["0x0d"] = 13; hexadecimal_to_integer["0x0e"] = 14; hexadecimal_to_integer["0x0f"] = 15; hexadecimal_to_integer["0x10"] = 16; hexadecimal_to_integer["0x11"] = 17; hexadecimal_to_integer["0x12"] = 18; hexadecimal_to_integer["0x13"] = 19;
        hexadecimal_to_integer["0x14"] = 20; hexadecimal_to_integer["0x15"] = 21; hexadecimal_to_integer["0x16"] = 22; hexadecimal_to_integer["0x17"] = 23; hexadecimal_to_integer["0x18"] = 24; hexadecimal_to_integer["0x19"] = 25; hexadecimal_to_integer["0x1a"] = 26; hexadecimal_to_integer["0x1b"] = 27; hexadecimal_to_integer["0x1c"] = 28; hexadecimal_to_integer["0x1d"] = 29;
        hexadecimal_to_integer["0x1e"] = 30; hexadecimal_to_integer["0x1f"] = 31; hexadecimal_to_integer["0x20"] = 32; hexadecimal_to_integer["0x21"] = 33; hexadecimal_to_integer["0x22"] = 34; hexadecimal_to_integer["0x23"] = 35; hexadecimal_to_integer["0x24"] = 36; hexadecimal_to_integer["0x25"] = 37; hexadecimal_to_integer["0x26"] = 38; hexadecimal_to_integer["0x27"] = 39;
        hexadecimal_to_integer["0x28"] = 40; hexadecimal_to_integer["0x29"] = 41; hexadecimal_to_integer["0x2a"] = 42; hexadecimal_to_integer["0x2b"] = 43; hexadecimal_to_integer["0x2c"] = 44; hexadecimal_to_integer["0x2d"] = 45; hexadecimal_to_integer["0x2e"] = 46; hexadecimal_to_integer["0x2f"] = 47; hexadecimal_to_integer["0x30"] = 48; hexadecimal_to_integer["0x31"] = 49;
        hexadecimal_to_integer["0x32"] = 50; hexadecimal_to_integer["0x33"] = 51; hexadecimal_to_integer["0x34"] = 52; hexadecimal_to_integer["0x35"] = 53; hexadecimal_to_integer["0x36"] = 54; hexadecimal_to_integer["0x37"] = 55; hexadecimal_to_integer["0x38"] = 56; hexadecimal_to_integer["0x39"] = 57; hexadecimal_to_integer["0x3a"] = 58; hexadecimal_to_integer["0x3b"] = 59;
        hexadecimal_to_integer["0x3c"] = 60; hexadecimal_to_integer["0x3d"] = 61; hexadecimal_to_integer["0x3e"] = 62; hexadecimal_to_integer["0x3f"] = 63; hexadecimal_to_integer["0x40"] = 64; hexadecimal_to_integer["0x41"] = 65; hexadecimal_to_integer["0x42"] = 66; hexadecimal_to_integer["0x43"] = 67; hexadecimal_to_integer["0x44"] = 68; hexadecimal_to_integer["0x45"] = 69;
        hexadecimal_to_integer["0x46"] = 70; hexadecimal_to_integer["0x47"] = 71; hexadecimal_to_integer["0x48"] = 72; hexadecimal_to_integer["0x49"] = 73; hexadecimal_to_integer["0x4a"] = 74; hexadecimal_to_integer["0x4b"] = 75; hexadecimal_to_integer["0x4c"] = 76; hexadecimal_to_integer["0x4d"] = 77; hexadecimal_to_integer["0x4e"] = 78; hexadecimal_to_integer["0x4f"] = 79;
        hexadecimal_to_integer["0x50"] = 80; hexadecimal_to_integer["0x51"] = 81; hexadecimal_to_integer["0x52"] = 82; hexadecimal_to_integer["0x53"] = 83; hexadecimal_to_integer["0x54"] = 84; hexadecimal_to_integer["0x55"] = 85; hexadecimal_to_integer["0x56"] = 86; hexadecimal_to_integer["0x57"] = 87; hexadecimal_to_integer["0x58"] = 88; hexadecimal_to_integer["0x59"] = 89;
        hexadecimal_to_integer["0x5a"] = 90; hexadecimal_to_integer["0x5b"] = 91; hexadecimal_to_integer["0x5c"] = 92; hexadecimal_to_integer["0x5d"] = 93; hexadecimal_to_integer["0x5e"] = 94; hexadecimal_to_integer["0x5f"] = 95; hexadecimal_to_integer["0x60"] = 96; hexadecimal_to_integer["0x61"] = 97; hexadecimal_to_integer["0x62"] = 98; hexadecimal_to_integer["0x63"] = 99;
        hexadecimal_to_integer["0x64"] = 100; hexadecimal_to_integer["0x65"] = 101; hexadecimal_to_integer["0x66"] = 102; hexadecimal_to_integer["0x67"] = 103; hexadecimal_to_integer["0x68"] = 104; hexadecimal_to_integer["0x69"] = 105; hexadecimal_to_integer["0x6a"] = 106; hexadecimal_to_integer["0x6b"] = 107; hexadecimal_to_integer["0x6c"] = 108; hexadecimal_to_integer["0x6d"] = 109;
        hexadecimal_to_integer["0x6e"] = 110; hexadecimal_to_integer["0x6f"] = 111; hexadecimal_to_integer["0x70"] = 112; hexadecimal_to_integer["0x71"] = 113; hexadecimal_to_integer["0x72"] = 114; hexadecimal_to_integer["0x73"] = 115; hexadecimal_to_integer["0x74"] = 116; hexadecimal_to_integer["0x75"] = 117; hexadecimal_to_integer["0x76"] = 118; hexadecimal_to_integer["0x77"] = 119;
        hexadecimal_to_integer["0x78"] = 120; hexadecimal_to_integer["0x79"] = 121; hexadecimal_to_integer["0x7a"] = 122; hexadecimal_to_integer["0x7b"] = 123; hexadecimal_to_integer["0x7c"] = 124; hexadecimal_to_integer["0x7d"] = 125; hexadecimal_to_integer["0x7e"] = 126; hexadecimal_to_integer["0x7f"] = 127; hexadecimal_to_integer["0x80"] = 128; hexadecimal_to_integer["0x81"] = 129;
        hexadecimal_to_integer["0x82"] = 130; hexadecimal_to_integer["0x83"] = 131; hexadecimal_to_integer["0x84"] = 132; hexadecimal_to_integer["0x85"] = 133; hexadecimal_to_integer["0x86"] = 134; hexadecimal_to_integer["0x87"] = 135; hexadecimal_to_integer["0x88"] = 136; hexadecimal_to_integer["0x89"] = 137; hexadecimal_to_integer["0x8a"] = 138; hexadecimal_to_integer["0x8b"] = 139;
        hexadecimal_to_integer["0x8c"] = 140; hexadecimal_to_integer["0x8d"] = 141; hexadecimal_to_integer["0x8e"] = 142; hexadecimal_to_integer["0x8f"] = 143; hexadecimal_to_integer["0x90"] = 144; hexadecimal_to_integer["0x91"] = 145; hexadecimal_to_integer["0x92"] = 146; hexadecimal_to_integer["0x93"] = 147; hexadecimal_to_integer["0x94"] = 148; hexadecimal_to_integer["0x95"] = 149;
        hexadecimal_to_integer["0x96"] = 150; hexadecimal_to_integer["0x97"] = 151; hexadecimal_to_integer["0x98"] = 152; hexadecimal_to_integer["0x99"] = 153; hexadecimal_to_integer["0x9a"] = 154; hexadecimal_to_integer["0x9b"] = 155; hexadecimal_to_integer["0x9c"] = 156; hexadecimal_to_integer["0x9d"] = 157; hexadecimal_to_integer["0x9e"] = 158; hexadecimal_to_integer["0x9f"] = 159;
        hexadecimal_to_integer["0xa0"] = 160; hexadecimal_to_integer["0xa1"] = 161; hexadecimal_to_integer["0xa2"] = 162; hexadecimal_to_integer["0xa3"] = 163; hexadecimal_to_integer["0xa4"] = 164; hexadecimal_to_integer["0xa5"] = 165; hexadecimal_to_integer["0xa6"] = 166; hexadecimal_to_integer["0xa7"] = 167; hexadecimal_to_integer["0xa8"] = 168; hexadecimal_to_integer["0xa9"] = 169;
        hexadecimal_to_integer["0xaa"] = 170; hexadecimal_to_integer["0xab"] = 171; hexadecimal_to_integer["0xac"] = 172; hexadecimal_to_integer["0xad"] = 173; hexadecimal_to_integer["0xae"] = 174; hexadecimal_to_integer["0xaf"] = 175; hexadecimal_to_integer["0xb0"] = 176; hexadecimal_to_integer["0xb1"] = 177; hexadecimal_to_integer["0xb2"] = 178; hexadecimal_to_integer["0xb3"] = 179;
        hexadecimal_to_integer["0xb4"] = 180; hexadecimal_to_integer["0xb5"] = 181; hexadecimal_to_integer["0xb6"] = 182; hexadecimal_to_integer["0xb7"] = 183; hexadecimal_to_integer["0xb8"] = 184; hexadecimal_to_integer["0xb9"] = 185; hexadecimal_to_integer["0xba"] = 186; hexadecimal_to_integer["0xbb"] = 187; hexadecimal_to_integer["0xbc"] = 188; hexadecimal_to_integer["0xbd"] = 189;
        hexadecimal_to_integer["0xbe"] = 190; hexadecimal_to_integer["0xbf"] = 191; hexadecimal_to_integer["0xc0"] = 192; hexadecimal_to_integer["0xc1"] = 193; hexadecimal_to_integer["0xc2"] = 194; hexadecimal_to_integer["0xc3"] = 195; hexadecimal_to_integer["0xc4"] = 196; hexadecimal_to_integer["0xc5"] = 197; hexadecimal_to_integer["0xc6"] = 198; hexadecimal_to_integer["0xc7"] = 199;
        hexadecimal_to_integer["0xc8"] = 200; hexadecimal_to_integer["0xc9"] = 201; hexadecimal_to_integer["0xca"] = 202; hexadecimal_to_integer["0xcb"] = 203; hexadecimal_to_integer["0xcc"] = 204; hexadecimal_to_integer["0xcd"] = 205; hexadecimal_to_integer["0xce"] = 206; hexadecimal_to_integer["0xcf"] = 207; hexadecimal_to_integer["0xd0"] = 208; hexadecimal_to_integer["0xd1"] = 209;
        hexadecimal_to_integer["0xd2"] = 210; hexadecimal_to_integer["0xd3"] = 211; hexadecimal_to_integer["0xd4"] = 212; hexadecimal_to_integer["0xd5"] = 213; hexadecimal_to_integer["0xd6"] = 214; hexadecimal_to_integer["0xd7"] = 215; hexadecimal_to_integer["0xd8"] = 216; hexadecimal_to_integer["0xd9"] = 217; hexadecimal_to_integer["0xda"] = 218; hexadecimal_to_integer["0xdb"] = 219;
        hexadecimal_to_integer["0xdc"] = 220; hexadecimal_to_integer["0xdd"] = 221; hexadecimal_to_integer["0xde"] = 222; hexadecimal_to_integer["0xdf"] = 223; hexadecimal_to_integer["0xe0"] = 224; hexadecimal_to_integer["0xe1"] = 225; hexadecimal_to_integer["0xe2"] = 226; hexadecimal_to_integer["0xe3"] = 227; hexadecimal_to_integer["0xe4"] = 228; hexadecimal_to_integer["0xe5"] = 229;
        hexadecimal_to_integer["0xe6"] = 230; hexadecimal_to_integer["0xe7"] = 231; hexadecimal_to_integer["0xe8"] = 232; hexadecimal_to_integer["0xe9"] = 233; hexadecimal_to_integer["0xea"] = 234; hexadecimal_to_integer["0xeb"] = 235; hexadecimal_to_integer["0xec"] = 236; hexadecimal_to_integer["0xed"] = 237; hexadecimal_to_integer["0xee"] = 238; hexadecimal_to_integer["0xef"] = 239;
        hexadecimal_to_integer["0xf0"] = 240; hexadecimal_to_integer["0xf1"] = 241; hexadecimal_to_integer["0xf2"] = 242; hexadecimal_to_integer["0xf3"] = 243; hexadecimal_to_integer["0xf4"] = 244; hexadecimal_to_integer["0xf5"] = 245; hexadecimal_to_integer["0xf6"] = 246; hexadecimal_to_integer["0xf7"] = 247; hexadecimal_to_integer["0xf8"] = 248; hexadecimal_to_integer["0xf9"] = 249;
        hexadecimal_to_integer["0xfa"] = 250; hexadecimal_to_integer["0xfb"] = 251; hexadecimal_to_integer["0xfc"] = 252; hexadecimal_to_integer["0xfd"] = 253; hexadecimal_to_integer["0xfe"] = 254; hexadecimal_to_integer["0xff"] = 255; 
    }

    public bool xbmBool(int i, int k) {
        k -= 1;
        return xbm_bool_array[i][k];
    }

    public KunCode readXbm(char[][] args) {
        bool csv = false;
        bool kun_shrink = false;
        bool remove_first = true;
        bool sharp_symbols = false;
        const int CSV = 0;
        const int KUN = 1;
        const int KOPIC_EXTENSION_TEXT = 2;
        const int XBM = 3;
        int decrement_according_to_file = 0;
        int count = args.length;
        int k, u;
        Numbered n;
        char[] argument;
        char[] fileName = KOPIC_XBM;

        try {
            Stream buffered = new BufferedFile(MANAGE_TEXT); 
            char[] s;

            s = buffered.readLine();
            if(s != null) {
                decrement_according_to_file = atoi(s);

                s = buffered.readLine();
                if(s != null) {
                    cell_width = atoi(s);

                    s = buffered.readLine();
                    if(s != null) {
                        cell_height = atoi(s);
                    }

                }

           }

            buffered.close();
        } catch(OpenException e18) {
            error = "OpenException e18";
            createManageText();
        } catch(ReadException e11) {
            error = "ReadException e11";
            createManageText();
        }

        kc = new KunCode();
        kc.decrement_by_n = decrement_according_to_file;
        kc.defectiveCsv = true;
        kc.from_not_38 = false;
        kc.newXbm = false;

        int d;
        for(int i = 1; i < count; ++i) {
            argument = args[i];

            if(CSV_FILE == argument) {
                csv = true;
            } else if(EASY_TO_SEE == argument) {
                sharp_symbols = true;
            } else if(FROM_NOT_38 == argument) {
                kc.from_not_38 = true; // Use From, not 38.
            } else if(KUN_FILE == argument) {
                kun_shrink = true;
            } else if(NEW_XBM == argument) {
                kc.newXbm = true;
            } else if(NO_REMOVING == argument) {
                remove_first = false;
            } else if(version_guide == argument) {
                guide(WITHOUT_COMMENTS);
                kc.code = VERSION_GUIDE;
                return kc;
            } else if(version_comments == argument) {
                guide(COMMENTS);
                kc.code = VERSION_COMMENTS;
                return kc;
            } else if(XBM_RECOVERY == argument) {
                createKopicXbm();
                kc.code = RECOVERY;
                return kc;
            } else {
                d = find(argument, CSV_DECREMENT);
                if(d >= 0) {
                    n = new Numbered(argument);
                    kc.decrement_by_n = n.number();
                    if( CSV_DECREMENT == n.parameter() ) {
                        csv = true;
                        continue;
                    }

                    kc.decrement_by_n = 0;
                }

                d = find(argument, SIZE);
                if(d >= 0) {
                    n = new Numbered(argument);
                    cell_height = n.number();
                    if( n.trimmed_sequence() ) {
                        cell_width = n.subnumber();
                        if( SIZE == n.parameter() ) {
                            resize = true;
                            continue;
                        }

                    }

                    cell_height = CELL_HEIGHT;
                    cell_width = CELL_WIDTH;
                }

                fileName = argument;
            }

        }

        kc.cell_height = cell_height;
        kc.cell_width = cell_width;
        if(remove_first) {
            removeTemporaryFiles();
        }

        try {
            Stream test = new BufferedFile(FALSE_TEXT); 
            test.close();
        } catch(OpenException e8) {
            error = "OpenException e8";
            createFalseText();
        }

        try {
            Stream test = new BufferedFile(TRUE_TEXT); 
            test.close();
        } catch(OpenException e7) {
            error = "OpenException e7";
            createTrueText();
        }

        kc.file_name = fileName;
        int extension = checkExtension(kc.file_name);
        switch(extension) {
        case CSV:
            kc.defectiveCsv = true;
            kc.code = CSV_SPECIFIED;
            return kc;
        case KUN:
            if( kunExpand(kc.file_name) ) {
                kc.code = KUN_EXPAND;
                return kc;
            }

            kc.code = XBM_FILE_ERROR;
            return kc;
        case KOPIC_EXTENSION_TEXT:
            if( itemsTreatedAsTrue(kc.file_name) ) { // Replacing true caracters with quotes to produce a temporary file (kc.trueKopicText)
                if( itemsTreatedAsFalse(kc.trueKopicText) ) {
                    kc.code = TREATED_AS_FALSE;
                    return kc;
                }

            } else {
                kc.code = NOT_TREATED_AS_TRUE;
                return kc;
            }

            kc.code = XBM_FILE_ERROR;
            return kc;
        default:
            break;
        }

        if(! kc.defectiveCsv) {
// Regardless of kopic.xbm, software preserves kopic.csv because it is perfect.
            kc.file_name = KOPIC_CSV;
            kc.code = CSV_SPECIFIED;
            return kc;
        }

        try {
            Stream buffered = new BufferedFile(kc.file_name); 
            char[] s;
            int q;
            for(;;) {
// Reading height and width
                s = buffered.readLine();
                if(s == null) {
                    break;
                }

                q = find(s, "w");
                if(q >= 0) {
                    u = sequence_of_non_numbers(s);
                    kc.data_width = atoi(s[u..$]);
                } else {
                    q = find(s, "g");
                    if(q >= 0) {
                        u = sequence_of_non_numbers(s);
                        kc.data_height = atoi(s[u..$]);
                    } else {
                        break;
                    }

                }

            }

// Reading bits to produce arrays 
            k = 0; // data_bits[k]
            char[] head, tail;

            for(;;) {
                s = buffered.readLine();
                if(s == null) {
                    break;
                }

                q = find(s, "s");
                if(q >= 0) {
                    continue;
                } else {
                    tail = s ~ newline;

                    for(;;) {
                        q = find(tail, ZERO_X);
                        if(q >= 0) {
                            ;
                        } else {
                            break;
                        }

                        tail = tail[q..$];
                        q = find(tail, COMMA);
                        if(q >= 0) {
                            ;
                        } else {
                            q = find(tail, newline);
                            if(q >= 0) {
                                ;
                            } else {
                                break;
                            }

                        }

                        head = tail[0..q];
                        tail = tail[q..$];
                        data_bits[k++] = hexadecimal_to_integer[head];
                    }

                }

            }

            buffered.close();
        } catch(OpenException e19) {
            error = "OpenException e19";
            createErrorText(kc.file_name);
            kc.code = XBM_FILE_ERROR;
            return kc;
        } catch(ReadException e10) {
            error = "ReadException e10";
            createErrorText(kc.file_name);
            kc.code = XBM_FILE_ERROR;
            return kc;
        }

        try {
            Stream buffered = new BufferedFile(KOPIC_CSV); 
            buffered.close();
        } catch(OpenException e9) {
            error = "OpenException e9";
            createErrorText("Software may produce kopic.csv from kopic.xbm");
            csv = true;
        }

        data_height = kc.data_height;
        data_width = kc.data_width;
        decrement_by_n = kc.decrement_by_n;
        xbmBooleanArray(kc.data_width, kc.data_height);

        if(csv) {
            createKopicCsv( standardTerms(), kc.data_width, kc.data_height, kc.from_not_38);
            kc.code = XBM_TO_CSV;
            return kc;
        }

        if(kun_shrink) {
            shrink(kc.data_width, kc.data_height);
            kc.code = KUN_SHRINK;
            return kc;
        }

        if(sharp_symbols) {
            sharp_text(KOPIC_TEXT, kc.data_width, kc.data_height);
            kc.code = SHARP_TEXT;
            return kc;
        }

        createKopicCsv( inheritTerms(), kc.data_width, kc.data_height, kc.from_not_38);
        kc.code = XBM_READ;
        return kc;
    }

    public char[] inquiry() {
        return error;
    }

}

